def init_cond(well):
    """
    Generates the temperature profiles at time 0, before starting the operation.
    :param well: a well object created from the function set_well()
    :return: object with initial temperature profiles
    """

    # Initial Conditions
    Tfto = [well.ts]   # Temperature of the fluid inside the tubing at RKB
    Tto = [well.ts]    # Temperature of the tubing at RKB, t=0
    Tao = [well.ts]      # Temperature of the fluid inside the annulus at RKB, t=0
    Tco = [well.ts]    # Temperature of the casing at RKB, t=0
    Tsro = [well.ts]    # Temperature of the surrounding space at RKB, t=0
    Tfm = [well.ts]      # Temperature of the formation at RKB

    for j in range(1, well.zstep):

        if j <= well.riser:
            Tg = well.wtg    # Water Thermal Gradient for the Riser section
        else:
            Tg = well.gt      # Geothermal Gradient below the Riser section

        deltaT = Tsro[j - 1] + Tg*(well.tvd[j]-well.tvd[j-1])/well.deltaz

        # Generating the Temperature Profile at t=0
        Tfto.append(deltaT)
        Tto.append(deltaT)
        Tao.append(deltaT)
        Tco.append(deltaT)
        Tsro.append(deltaT)
        Tfm.append(deltaT)

    class InitCond(object):
        def __init__(self):
            self.tfto = Tfto
            self.tto = Tto
            self.tao = Tao
            self.tco = Tco
            self.tsro = Tsro
            self.tfm = Tfm

    return InitCond()



def initial_density(well, initcond, section='tubing'):
    """
    Function to calculate the density profile for the first time step
    :param section: 'tubing' or 'annular'
    :param well: a well object created from the function set_well()
    :param initcond: a initial conditions object with the formation temperature profile
    :return: the density profile and the initial density at surface conditions
    """

    if section == 'tubing':
        beta = well.beta
        alpha = well.alpha
        rho = well.rhof

    if section == 'annular':
        beta = well.beta_a
        alpha = well.alpha_a
        rho = well.rhof_a

    rhof_initial = rho
    pressure = [rho * 9.81 * i for i in well.tvd]
    rhof = [rhof_initial * (1 + (x - 10 ** 5) / beta - alpha * (y - well.ts)) for x, y in
            zip(pressure, initcond.tfto)]
    pressure = [x * 9.81 * y for x, y in zip(rhof, well.tvd)]

    rhof = [rhof_initial * (1 + (x - 10 ** 5) / beta - alpha * (y - well.ts)) for x, y in
            zip(pressure, initcond.tfto)]

    return rhof, rhof_initial


def calc_density(well, initcond, rhof_initial, section='tubing'):
    """
    Function to calculate the density profile
    :param section: 'tubing' or 'annular'
    :param well: a well object created from the function set_well()
    :param initcond: a initial conditions object with the formation temperature profile
    :param rhof_initial: initial density at surface conditions
    :param flow: boolean to define if the section is flowing
    :return: density profile
    """

    if section == 'tubing':
        beta = well.beta
        alpha = well.alpha
        flow = True
        temp = initcond.tfto
        rho = well.rhof
    if section == 'annular':
        beta = well.beta_a
        alpha = well.alpha_a
        flow = False
        temp = initcond.tao
        rho = well.rhof_a

    pressure_h = [x * 9.81 * y for x, y in zip(rho, well.tvd)]

    if flow:
        pressure_f = [x * (well.md[-1] / well.dti) * (1/2) * y * well.vp **2 for x, y in zip(well.f_p, rho)]
    else:
        pressure_f = [0] * len(well.md)

    pressure = [x + y for x, y in zip(pressure_h, pressure_f)]

    rhof = [rhof_initial * (1 + (x - 10 ** 5) / beta - alpha * (y - well.ts)) for x, y in
            zip(pressure, temp)]

    return rhof


def calc_vicosity(api, initcond):
    """
    Function to calculate the viscosity profile (Glaso, 1980)
    :param api: API gravity of the produced hydrocarbon
    :param initcond: a initial conditions object with the formation temperature profile
    :return: viscosity profile
    """

    from math import log10

    visc_t, visc_a = [], []
    for x, y in zip(initcond.tfto, initcond.tao):
        x = (x * 1.8) + 32      # converting °C to °F
        y = (y * 1.8) + 32      # converting °C to °F
        c = 3.141 * (10 ** 10) * (x ** (-3.444))
        d = (10.313 * log10(x)) - 36.447
        visc_t.append((c * log10(api) ** d)/1000)
        c = 3.141 * (10 ** 10) * (y ** (-3.444))
        d = (10.313 * log10(y)) - 36.447
        visc_a.append((c * log10(api) ** d)/1000)

    return visc_t, visc_a


from .inputs import inputs_dict
from .well_system import set_well
from .linearsystem import calc_temperature_distribution
from .plot import plot_behavior
from scipy.interpolate import make_interp_spline
import numpy as np
import scipy.signal


def calc_temp(trajectory, casings=None, set_inputs=None, operation='drilling', time_steps=210, smooth=True,
              cells_no=None):
    """
    Function to calculate the well temperature distribution during a specific operation at a certain time.

    Arguments:
        trajectory: (obj) wellbore trajectory object
        casings: list of dictionaries with casings characteristics (od, id and depth)
        set_inputs: dictionary with parameters to set.
        operation: define operation type. ('drilling', 'circulating')
        time_steps: number of time steps to run calculations.
        smooth: smooth the temperature profiles.
        cells_no: (int) number of cells. If None -> keep same number of cells than trajectory

    Returns:
        Well temperature distribution object
    """

    tdata = inputs_dict(casings)

    if set_inputs is not None:
        for x in set_inputs:  # changing default values
            if x in tdata:
                tdata[x] = set_inputs[x]
            else:
                raise TypeError('%s is not a parameter' % x)

    md_initial = tdata['water_depth']
    md_final = trajectory.md[-1]

    well = set_well(tdata, trajectory, operation, cells_no)

    time = []
    rop_steps = []
    depths = sorted([x[2] for x in well.casings])
    if operation == 'drilling':

        prev_point = md_initial

        cummulative_time = []

        for x in range(len(depths)):
            distance = depths[x] - prev_point
            time_section = distance / well.rop_list[x]
            time.append(time_section)
            cummulative_time.append(sum(time))
            prev_point = depths[x]
        time.append((md_final - prev_point) / well.rop_list[-1])
        cummulative_time.append(sum(time))
        depths = list(depths) + [md_final]
        if depths[0] == 0:
            depths = depths[1:]
            time = time[1:]
            cummulative_time = cummulative_time[1:]

        tcirc = sum(time) * 3600  # drilling time, s
        for x in cummulative_time:
            rop_steps.append(round(x * 3600 / tcirc / time_steps))
    else:
        tcirc = tdata['time'] * 3600    # circulating time, s

    time_step = tcirc / time_steps  # seconds per time step

    log_temp_values(well, initial=True)     # log initial temperature distribution
    well.delta_time = time_step
    time_n = time_step
    well.op = operation

    td = well.cells_no-1

    for x in range(time_steps):

        if well.op == 'drilling':
            d = depths[0]
            rop = well.rop_list[0]
            t = time[0] * 3600
            for y in range(len(time)):
                if time_n < sum(time[:y+1])*3600:
                    d = depths[y]
                    if len(well.rop_list) > 1:
                        rop = well.rop_list[y]
                    t = sum(time[:y+1])*3600
                    break

            bit_depth = d - rop/3600 * (t - time_n)
            bit_position = round(bit_depth / well.depth_step)
            td = bit_position

        if time_steps > 1:

            if td > 0:
                well = calc_temperature_distribution(well, time_step, td)
                well = define_temperatures(well, td)

                log_temp_values(well, time_n)
            well.time = time_n / 3600

            if well.op == 'drilling':
                if x in rop_steps:
                    well.temperatures['in_pipe'] = well.temp_fm
                    well.temperatures['pipe'] = well.temp_fm
                    well.temperatures['annulus'] = well.temp_fm
                    well.temperatures['casing'] = well.temp_fm
                    well.temperatures['riser'] = well.temp_fm
                    well.temperatures['sr'] = well.temp_fm
                    for i in well.sections:
                        for j in range(well.cells_no):
                            i[j]['temp'] = well.temp_fm[j]
                            i[j]['temp_fm'] = well.temp_fm[j]

        time_n += time_step

    if smooth:
        smooth_results(well)

    return well


def define_temperatures(well, bit_position):
    """
    Make the temperature values more reachable since they are is a dictionary along the entire well. Once this function
    takes place, the temperatures will be available as lists.
    :return: a dictionary with lists of temperature values and also a list with respective depth points.
    """

    temp_in_pipe = [x['temp'] for x in well.sections[0][:bit_position+1]] + \
                   [None] * (well.cells_no - (bit_position + 1))
    temp_pipe = [x['temp'] for x in well.sections[1][:bit_position+1]] + \
                [None] * (well.cells_no - (bit_position + 1))
    temp_annulus = [x['temp'] for x in well.sections[2][:bit_position+1]] + \
                   [None] * (well.cells_no - (bit_position + 1))
    temp_casing = []
    temp_riser = []
    temp_sr = [x['temp'] for x in well.sections[4]]
    for y, x in enumerate(well.md):

        if well.riser_cells > 0 and x < well.water_depth:
            temp_casing.append(None)
            temp_riser.append(well.sections[3][y]['temp'])
        else:
            temp_riser.append(None)
            if x <= well.casings[0, 2]:
                temp_casing.append(well.sections[3][y]['temp'])
            else:
                temp_casing.append(None)

    well.temperatures = {'md': well.md,
                         'formation': well.temp_fm,
                         'in_pipe': temp_in_pipe,
                         'pipe': temp_pipe,
                         'annulus': temp_annulus,
                         'casing': temp_casing,
                         'riser': temp_riser,
                         'sr': temp_sr}

    return well


def log_temp_values(well, time=0.0, initial=False):
    time = round(time/3600, 2)
    if initial:
        well.temp_log = [{'time': time,
                          'in_pipe': well.temp_fm,
                          'pipe': well.temp_fm,
                          'annulus': well.temp_fm,
                          'casing': well.temp_fm,
                          'riser': well.temp_fm,
                          'sr': well.temp_fm}]
    else:
        well.temp_log.append(
            {'time': time,
             'in_pipe': [x['temp'] for x in well.sections[0]],
             'pipe': [x['temp'] for x in well.sections[1]],
             'annulus': well.temperatures['annulus'],
             'casing': well.temperatures['casing'],
             'riser': well.temperatures['riser'],
             'sr': well.temperatures['sr']}
        )


def temperature_behavior(well):
    """
    Function to simulate the temperature behavior.

    Arguments:
        well (obj): well temperature distribution object

    Returns:
        temperature behavior object
    """

    time = []
    temp_bottom = []
    temp_outlet = []
    temp_max = []
    temp_fm = []

    for x in well.temp_log[1:]:
        cells = len(well.md) - x['annulus'].count(None)
        time.append(x['time'])
        temp_bottom.append(x['in_pipe'][cells-1])
        temp_outlet.append(x['annulus'][0])
        temp_max.append(max(x['annulus'][:cells-1]))
        temp_fm.append(well.temp_fm[cells-1])

    temp_bottom = list(scipy.signal.savgol_filter(temp_bottom, 59, 3))
    temp_max = list(scipy.signal.savgol_filter(temp_max, 59, 3))
    temp_outlet = list(scipy.signal.savgol_filter(temp_outlet, 59, 3))
    temp_fm = list(scipy.signal.savgol_filter(temp_fm, 59, 3))

    class TempBehavior(object):
        def __init__(self):
            self.time = time
            self.bottom = temp_bottom
            self.outlet = temp_outlet
            self.max = temp_max
            self.formation_td = temp_fm

        def plot(self, title=True):
            fig = plot_behavior(self, title)

            return fig

    return TempBehavior()


def smooth_results(well):

    well.temperatures['in_pipe'] = list(scipy.signal.savgol_filter(well.temperatures['in_pipe'], 15, 3))
    well.temperatures['annulus'] = list(scipy.signal.savgol_filter(well.temperatures['annulus'], 15, 2))

    cells = len(well.md) - np.count_nonzero(np.isnan(well.temperatures['annulus']))
    ref = int(0.9 * cells - well.time*0.2)
    t_bottom = np.mean(well.temperatures['annulus'][ref:])

    x_new = well.md[ref:]

    # Smooth annulus
    interp_ann = make_interp_spline([well.md[ref - 1], well.md[ref], well.md[-1]],
                                    [well.temperatures['annulus'][ref - 1], well.temperatures['annulus'][ref],
                                     t_bottom],
                                    k=2)
    temp_annulus = interp_ann(x_new)

    # Smooth in_pipe
    interp_in_pipe = make_interp_spline(well.md[:ref-2] + [well.md[-1]],
                                        well.temperatures['in_pipe'][:ref-2] + [t_bottom],
                                        k=2)
    temp_in_pipe = interp_in_pipe(x_new)

    well.temperatures['in_pipe'] = well.temperatures['in_pipe'][:ref] + list(temp_in_pipe)
    well.temperatures['annulus'] = well.temperatures['annulus'][:ref] + list(temp_annulus)



from math import pi, log


def add_heat_coefficients(well, delta_time, bit_position):
    for x, y in enumerate(well.sections[0][:bit_position+1]):
        y['comp_N/S'] = ((y['rho'] * y['shc'] * well.vp) / well.depth_step) / 2
        y['comp_E'] = (2 * well.h1[x] / well.r1) / 2
        y['comp_HeatSource'] = calc_heat_source(well, well.torque[x], y['f'], y['rho'], case='pipe',
                                                operation=well.op)
        y['comp_time'] = y['rho'] * y['shc'] / delta_time
    for x, y in enumerate(well.sections[1][:bit_position+1]):
        y['comp_N/S'] = (y['tc'] / (well.depth_step ** 2)) / 2
        y['comp_E'] = (2 * well.r2 * well.h2[x] / ((well.r2 ** 2) - (well.r1 ** 2))) / 2
        y['comp_W'] = (2 * well.r1 * well.h1[x] / ((well.r2 ** 2) - (well.r1 ** 2))) / 2
        y['comp_time'] = y['rho'] * y['shc'] / delta_time
    for x, y in enumerate(well.sections[2][:bit_position+1]):
        y['comp_N/S'] = (y['rho'] * y['shc'] * well.va / well.depth_step) / 2
        y['comp_E'] = (2 * well.annular_or * well.h3[x] / ((well.annular_or ** 2) - (well.r2 ** 2))) / 2
        y['comp_W'] = (2 * well.r2 * well.h2[x] / ((well.annular_or ** 2) - (well.r2 ** 2))) / 2
        y['comp_HeatSource'] = calc_heat_source(well, well.torque[x], y['f'], y['rho'], case='annular',
                                                operation=well.op)
        y['comp_time'] = y['rho'] * y['shc'] / delta_time
    for x, y in enumerate(well.sections[3][:bit_position+1]):
        y['comp_N/S'] = (y['tc'] / (well.depth_step ** 2)) / 2
        y['comp_E'] = (2 * y['tc'] / ((well.sr_ir ** 2) - (well.annular_or ** 2))) / 2
        y['comp_W'] = (2 * well.annular_or * well.h3[x] / ((well.sr_ir ** 2) - (well.annular_or ** 2))) / 2
        y['comp_time'] = y['rho'] * y['shc'] / delta_time
    for x, y in enumerate(well.sections[4][:bit_position+1]):
        y['comp_N/S'] = (y['tc'] / (well.depth_step ** 2)) / 2
        y['comp_E'] = (y['tc'] / (well.sr_or * (well.sr_or - well.sr_ir) * log(well.sr_or / well.sr_ir))) / 2
        y['comp_W'] = (y['tc'] / (well.sr_or * (well.sr_or - well.sr_ir) * log(well.fm_rad / well.sr_or))) / 2
        y['comp_time'] = y['rho'] * y['shc'] / delta_time

    if well.op == 'drilling':
        # heat coefficients at drill bit
        joule = 4.1868  # Joule's constant  [Nm/cal]
        bit_cell = well.sections[0][bit_position]
        q_bit = (1 / joule) * (1 - well.bit_n) * (well.wob * (well.rop / 3600) + 2 * pi * (well.rpm / 60) * well.tbit) \
            + 0.7 * (well.q / 3600) * (bit_cell['rho'] / (2 * 9.81)) * ((well.q / 3600) / (0.95 * well.an)) ** 2
        v_bit = well.q / well.an
        bit_cell['comp_N/S'] = ((bit_cell['rho'] * bit_cell['shc'] * v_bit) / well.depth_step) / 2
        bit_cell['comp_E'] = (2 * well.h1[bit_position] / well.annular_or) / 2  # East component
        bit_cell['comp_HeatSource'] = q_bit / well.an  # Heat source term
        bit_cell['comp_time'] = bit_cell['rho'] * bit_cell['shc'] / delta_time  # Time component


def calc_heat_source(well, torque, f, rho_fluid, case='pipe', operation='drilling'):

    heat_source_term = 0

    if operation == 'drilling':
        if case == 'pipe':
            qp = 2 * pi * (well.rpm / 60) * torque
            + 0.2 * well.q * 2 * f * rho_fluid * (well.vp ** 2) * (well.md[-1] / (well.pipe_id * 127.094 * 10 ** 6))
            heat_source_term = qp / (pi * (well.r1 ** 2))

        else:
            qa = (0.085 * (2 * well.k * well.md[-1] / ((well.annular_or - well.r2) * (127.094 * 10 ** 6))) *
                  ((2 * (well.n + 1) * well.q) / (well.n * pi * (well.annular_or + well.r2) *
                                                  (well.annular_or - well.r2) ** 2)) ** well.n) * (
                             1 + (3 / 2) * well.dp_e ** 2) ** 0.5
            heat_source_term = qa / (pi * ((well.annular_or ** 2) - (well.r2 ** 2)))

    return heat_source_term



import numpy as np


def temp_time(n, well, log=True, units='metric', time_delta=900):
    """
    Function to calculate the well temperature distribution during certain production time (n)
    :param n: production time, hours
    :param well: a well object created with the function set_well() from input.py
    :param log: save distributions between initial time and circulation time n (each 1 hour)
    :param units: system of units ('metric' or 'english')
    :param time_delta: duration of each time step (seconds)
    :return: a well temperature distribution object
    """
    from .initcond import init_cond
    from .heatcoefficients import heat_coef
    from .linearsystem import temp_calc
    from ..plot import profile
    from math import log, nan
    # Simulation main parameters
    time = n  # circulating time, h
    tcirc = time * 3600  # circulating time, s
    deltat = 60 * time
    if type(time_delta) == int:
        deltat = time_delta
    tstep = int(tcirc / deltat)
    ic = init_cond(well)
    tfm = ic.tfm
    tt = ic.tto
    t3 = ic.tco

    well = well.define_viscosity(ic)
    well = well.define_density(ic, cond=0)

    hc = heat_coef(well, deltat, tt, t3)
    temp = temp_calc(well, ic, hc)
    temp.tft = temp.tt = temp.ta = temp.t3 = temp.tsr = tfm
    for x in range(len(tfm)):
        if temp.tc[x] != nan:
            temp.tc[x] = tfm[x]
        if temp.tr[x] != nan:
            temp.tr[x] = tfm[x]
        if temp.toh[x] != nan:
            temp.toh[x] = tfm[x]

    temp_initial = temp
    temp_initial.tft = ic.tfm
    temp_initial.tt = ic.tfm
    temp_initial.ta = ic.tfm

    temp_log = [temp_initial, temp]
    time_log = [0, deltat / 3600]

    for x in range(tstep-1):
        well = well.define_viscosity(ic)
        well = well.define_density(ic, cond=1)

        ic.tfto = temp.tft
        ic.tto = temp.tt
        ic.tao = temp.ta
        ic.tco = temp.t3
        ic.tsr = temp.tsr
        hc_new = heat_coef(well, deltat, ic.tto, ic.tco)
        temp = temp_calc(well, ic, hc_new)

        if units == 'english':
            temp.tft_output = [(i/(5/9)+32) for i in temp.tft]
            temp.tt_output = [(i/(5/9)+32) for i in temp.tt]
            temp.ta_output = [(i/(5/9)+32) for i in temp.ta]
            temp.tc_output = [(i/(5/9)+32) for i in temp.tc if type(i) == np.float64]
            temp.tr_output = [(i/(5/9)+32) for i in temp.tr if type(i) == np.float64]
            temp.tsr_output = [(i/(5/9)+32) for i in temp.tsr]
            temp.md_output = [i*3.28 for i in well.md]

        if log:
            temp_log.append(temp)
            time_log.append(time_log[-1] + time_log[1])

    if units == 'english':
        temp.tft = temp.tft_output
        temp.tt = temp.tt_output
        temp.ta = temp.ta_output
        temp.tc = temp.tc_output
        temp.tr = temp.tr_output
        temp.sr = temp.tsr_output
        temp.md = temp.md_output
        tfm = [(i / (5 / 9) + 32) for i in tfm]

    class TempDist(object):
        def __init__(self):
            self.tft = temp.tft
            self.tt = temp.tt
            self.ta = temp.ta
            self.tc = temp.tc
            self.tr = temp.tr
            self.tsr = temp.tsr
            self.tfm = tfm
            self.time = time
            self.md = well.md
            self.riser = well.riser
            self.deltat = deltat
            self.csgs_reach = temp.csgs_reach
            if log:
                self.temp_log = temp_log
                self.time_log = time_log

        def well(self):
            return well

        def plot(self):
            return profile(self, units, operation='production')

        def behavior(self):
            temp_behavior_production = temp_behavior(self)
            return temp_behavior_production

    return TempDist()


def temp_behavior(temp_dist):

    ta = [x.ta for x in temp_dist.temp_log]

    tout = []

    for n in range(len(ta)):
        tout.append(ta[n][0])

    class Behavior(object):
        def __init__(self):
            self.finaltime = temp_dist.time
            self.tout = tout
            self.tfm = temp_dist.tfm
            self.time = temp_dist.time_log

        def plot(self):
            from ..plot import behavior
            return behavior(self, operation='production')

    return Behavior()


def temp(well_trajectory, n, casings=[], d_openhole=0.216, change_input={}, log=False, units='metric', time_delta=900):
    """
    Main function to calculate the well temperature distribution during production operation. This function allows to
    set the wellpath and different parameters involved.
    :param well_trajectory: wellbore trajectory object
    :param n: production time, hours
    :param casings: list of dictionaries with casings characteristics (od, id and depth)
    :param change_input: dictionary with parameters to set.
    :param log: save distributions between initial time and circulation time n (each 1 hour)
    :param units: system of units ('metric' or 'english')
    :param time_delta: duration of each time step (seconds)
    :return: a well temperature distribution object
    """
    from .input import data, set_well

    tdata = data(casings, d_openhole, units)
    for x in change_input:   # changing default values
        if x in tdata:
            tdata[x] = change_input[x]
        else:
            raise TypeError('%s is not a parameter' % x)

    well = set_well(tdata, well_trajectory, units)
    temp_distribution = temp_time(n, well, log, units, time_delta)

    return temp_distribution


def input_info(about='all'):
    from .input import info
    info(about)



import torque_drag as td
import numpy as np
from math import pi, log
from copy import deepcopy


def create_depth_cells(trajectory, cells_no=None):

    if cells_no is not None:
        md_new = list(np.linspace(0, max(trajectory.md), num=cells_no))
        tvd_new, inclination, azimuth = [], [], []
        for i in md_new:
            tvd_new.append(np.interp(i, trajectory.md, trajectory.tvd))
            inclination.append(np.interp(i, trajectory.md, trajectory.inclination))
            azimuth.append(np.interp(i, trajectory.md, trajectory.azimuth))
        depth_step = md_new[1]

        return md_new, tvd_new, depth_step, inclination, azimuth

    else:
        depth_step = trajectory.md[1] - trajectory.md[0]
        return trajectory.md, trajectory.tvd, depth_step, trajectory.inclination, trajectory.azimuth


def set_well(temp_dict, trajectory, operation, cells_no=None):
    q_conv = 60  # from m^3/min to m^3/h
    an_conv = 1 / 1500  # from in^2 to m^2
    diameter_conv = 0.0254  # from in to m

    class NewWell(object):
        def __init__(self):

            # DIMENSIONS
            self.trajectory = trajectory
            self.md, self.tvd, self.depth_step, self.inclination, self.azimuth = create_depth_cells(trajectory,
                                                                                                    cells_no)
            self.deltaz = self.depth_step
            self.cells_no = self.zstep = len(self.md)
            self.casings = temp_dict["casings"]  # casings array
            self.pipe_id = temp_dict["pipe_id"] * diameter_conv  # Drill String Inner  Diameter, m
            self.pipe_od = temp_dict["pipe_od"] * diameter_conv  # Drill String Outer Diameter, m
            # OFFSHORE
            self.water_depth = temp_dict["water_depth"]  # Water Depth, m
            self.riser_cells = round(self.water_depth / self.depth_step)  # number of grid cells for the riser
            self.riser_id = temp_dict["riser_id"] * diameter_conv  # Riser diameter Inner Diameter, m
            self.riser_od = temp_dict["riser_od"] * diameter_conv  # Riser diameter Outer Diameter, m
            self.th_grad_seawater = temp_dict["th_grad_seawater"]  # Seawater thermal grad., °C/cell
            # RADIUS (CALCULATED)
            self.r1 = self.pipe_id / 2  # Drill String Inner  Radius, m
            self.r2 = self.pipe_od / 2  # Drill String Outer Radius, m
            self.annular_or = self.casings[0, 1] / 2  # Casing Inner Radius, m
            self.riser_ir = self.riser_id / 2  # Riser Inner Radius, m
            self.riser_or = self.riser_od / 2  # Riser Outer Radius, m
            # Surrounding Space Outer Diameter, m
            self.sr_od = sorted([self.riser_od + 0.03, self.casings[-1, 0] + 0.03])[-1]
            self.fm_diam = temp_dict["fm_diam"] * diameter_conv  # Undisturbed Formation Diameter, m
            self.sr_diam = self.casings[0, 0]  # Surrounding Space Inner Diameter, m
            self.sr_ir = self.sr_diam / 2  # Surrounding Space Inner Radius m
            self.sr_or = self.sr_od / 2  # Surrounding Space Outer Radius, m
            self.sr_thickness = self.sr_or - self.sr_ir  # Surrounding thickness, m
            self.sr_fractions = get_fractions(self)
            self.fm_rad = self.fm_diam * diameter_conv / 2  # Undisturbed Formation Radius, m

            # OPERATIONAL
            self.temp_surface = temp_dict["temp_surface"]  # Surface Temperature (RKB), °C
            if temp_dict["temp_inlet"] is None:
                self.temp_inlet = self.temp_surface
            else:
                self.temp_inlet = temp_dict["temp_inlet"]  # Inlet Fluid temperature, °C
            self.th_grad_fm = temp_dict["th_grad_fm"]  # Geothermal gradient, from °C/m to °C/cell
            self.q = temp_dict["flowrate"] * q_conv  # Flow rate, m^3/h
            # Fluid velocity through the annular
            self.va = (self.q / (pi * ((self.annular_or ** 2) - (self.r2 ** 2)))) / 3600
            self.vp = (self.q / (pi * (self.r1 ** 2))) / 3600  # Fluid velocity through the drill pipe
            self.rpm = temp_dict["rpm"]  # Revolutions per minute
            self.tbit = temp_dict["tbit"]  # Torque on the bit, kN*m
            self.wob = temp_dict["wob"]  # Weight on bit, kN
            self.rop_list = temp_dict["rop"]
            self.rop = temp_dict["rop"][0]  # Rate of Penetration, m/h
            self.an = temp_dict["an"] * an_conv  # Area of the nozzles, m^2
            self.bit_n = temp_dict["bit_n"]  # drill bit efficiency
            self.dp_e = temp_dict["dp_e"]  # drill pipe eccentricity
            self.thao_o = temp_dict["thao_o"]
            self.k = temp_dict["k"]
            self.n = temp_dict["n"]
            self.visc = temp_dict['visc']

            # DENSITIES
            self.rho_fluid = temp_dict["rho_fluid"]  # Fluid
            self.rho_pipe = temp_dict["rho_pipe"]  # Drill Pipe
            self.rho_csg = temp_dict["rho_csg"]  # Casing
            self.rho_riser = temp_dict["rho_riser"]  # Riser
            self.rho_cem = temp_dict["rho_cem"]  # Cement Sheath
            self.rho_fm = temp_dict["rho_fm"]  # Formation
            self.rho_seawater = temp_dict["rho_seawater"]  # Seawater

            # Thermal conductivity  W/(m*°C)
            self.tc_fluid = temp_dict["tc_fluid"]  # Fluid
            self.tc_csg = temp_dict["tc_csg"]  # Casing
            self.tc_cem = temp_dict["tc_cem"]  # Cement
            self.tc_pipe = temp_dict["tc_pipe"]  # Drill Pipe
            self.tc_fm = temp_dict["tc_fm"]  # Formation
            self.tc_riser = temp_dict["tc_riser"]  # Riser
            self.tc_seawater = temp_dict["tc_seawater"]  # Seawater

            self.beta = temp_dict["beta"]  # isothermal bulk modulus, Pa
            self.alpha = temp_dict['alpha']  # Fluid Thermal Expansion Coefficient, 1/°C

            # Specific heat capacity, J/(kg*°C)
            self.shc_fluid = temp_dict["shc_fluid"]  # Fluid
            self.shc_csg = temp_dict["shc_csg"]  # Casing
            self.shc_cem = temp_dict["shc_cem"]  # Cement
            self.shc_pipe = temp_dict["shc_pipe"]  # Drill Pipe
            self.shc_riser = temp_dict["shc_riser"]  # Riser
            self.shc_seawater = temp_dict["shc_seawater"]  # Seawater
            self.shc_fm = temp_dict["shc_fm"]  # Formation

            self.sections = create_system(self)

            self.sections, self.temp_fm = calc_formation_temp(self)

            td_obj = td.calc(self,
                             dimensions={'od_pipe': self.pipe_od, 'id_pipe': self.pipe_id, 'length_pipe': self.md[-1],
                                         'od_annular': self.annular_or * 2},
                             case='static',
                             densities={'rhof': self.rho_fluid, 'rhod': self.rho_pipe},
                             fric=0.24, wob=self.wob, tbit=self.tbit, torque_calc=True)

            self.drag = td_obj.force['static']  # drag forces, kN
            self.torque = td_obj.torque['static']  # torque, kN*m

            self.sections, self.h1, self.h2, self.h3, self.h3r = extra_calcs(self)
            self.temperatures = None

            if operation == 'drilling':
                if len(self.casings) != len(self.rop_list) - 1 and self.casings[0, 2] != 0:
                    raise ValueError('ROP must be set for all the sections.')

    return NewWell()


def create_system(well):
    section_0, section_1, section_2, section_3, section_4 = [], [], [], [], []
    for x in range(well.cells_no):
        initial_dict = {'component': '', 'material': '', 'rho': 2.24, 'tc': '', 'shc': '',
                        'depth': well.md[x]}
        section_0.append(deepcopy(initial_dict))
        section_1.append(deepcopy(initial_dict))
        section_2.append(deepcopy(initial_dict))
        section_3.append(deepcopy(initial_dict))
        section_4.append(deepcopy(initial_dict))
    sections = [section_0, section_1, section_2, section_3, section_4]
    sections_names = ['section_0', 'section_1', 'section_2', 'section_3', 'section_4']

    for x, i in zip(sections, sections_names):
        for y in range(well.cells_no):
            x[y]['material'] = get_material(i, well.md[y],
                                            first_casing_depth=well.casings[0, 2],
                                            water_depth=well.water_depth)
            if x[y]['material'] == 'mixture':
                pipe_fraction, cement_fraction = get_fractions_at_depth(well, y)
                x[y]['rho'] = get_mixture_density(pipe_fraction, cement_fraction, well) * 1000
                x[y]['tc'] = get_mixture_thermal_conductivity(pipe_fraction, cement_fraction, well)
                x[y]['shc'] = get_mixture_heat_capacity(pipe_fraction, cement_fraction, well)

            else:
                x[y]['rho'] = get_density(x[y]['material'], well) * 1000
                x[y]['tc'] = get_thermal_conductivity(x[y]['material'], well)
                x[y]['shc'] = get_heat_capacity(x[y]['material'], well)
                if x[y]['material'] == 'fluid':
                    x[y]['visc'] = get_viscosity(well)

    return sections


def get_fractions_at_depth(well, cell):
    the_index = -1
    casing_depths = reversed([x[2] for x in well.casings])
    for x in casing_depths:
        if well.md[cell] > x:
            the_index -= 1

    pipe_fraction = well.sr_fractions[the_index][0]
    cement_fraction = well.sr_fractions[the_index][1]

    return pipe_fraction, cement_fraction


def get_density(material, well):
    rho = {'formation': well.rho_fm,
           'fluid': well.rho_fluid,
           'pipe': well.rho_fm,
           'casing': well.rho_csg,
           'riser': well.rho_riser,
           'seawater': well.rho_seawater,
           'cement': well.rho_cem}

    return rho[material]


def get_mixture_density(pipe_fraction, cement_fraction, well):
    formation_fraction = 1 - pipe_fraction - cement_fraction
    rho = pipe_fraction * get_density('casing', well) + \
        cement_fraction * get_density('cement', well) + \
        formation_fraction * get_density('formation', well)

    return rho


def get_viscosity(well):
    rho = {'fluid': well.visc}

    return rho['fluid']


def get_thermal_conductivity(material, well):
    tc = {'formation': well.tc_fm,
          'fluid': well.tc_fluid,
          'pipe': well.tc_fm,
          'casing': well.tc_csg,
          'riser': well.tc_riser,
          'seawater': well.tc_seawater,
          'cement': well.tc_cem}

    return tc[material]


def get_mixture_thermal_conductivity(pipe_fraction, cement_fraction, well):
    formation_fraction = 1 - pipe_fraction - cement_fraction
    rho = pipe_fraction * get_thermal_conductivity('casing', well) + \
        cement_fraction * get_thermal_conductivity('cement', well) + \
        formation_fraction * get_thermal_conductivity('formation', well)

    return rho


def get_heat_capacity(material, well):
    shc = {'formation': well.shc_fm,
           'fluid': well.shc_fluid,
           'pipe': well.shc_fm,
           'casing': well.shc_csg,
           'riser': well.shc_riser,
           'seawater': well.shc_seawater,
           'cement': well.shc_cem}

    return shc[material]


def get_mixture_heat_capacity(pipe_fraction, cement_fraction, well):
    formation_fraction = 1 - pipe_fraction - cement_fraction
    rho = pipe_fraction * get_heat_capacity('casing', well) + \
        cement_fraction * get_heat_capacity('cement', well) + \
        formation_fraction * get_heat_capacity('formation', well)

    return rho


def get_fractions(well):
    fractions = []
    for x in range(len(well.casings) - 1):
        pipe_thickness = (well.casings[x, 0] - well.casings[x, 1]) / 2
        pipe_fraction = pipe_thickness / well.sr_thickness
        cement_thickness = (well.casings[x + 1, 1] - well.casings[x, 0]) / 2
        cement_fraction = cement_thickness / well.sr_thickness
        fractions.append([pipe_fraction, cement_fraction])

    last_pipe_fraction = ((well.casings[-1, 0] - well.casings[-1, 1]) / 2) / well.sr_thickness
    last_cement_fraction = ((well.sr_or - well.casings[-1, 0]) / 2) / well.sr_thickness

    fractions.append([last_pipe_fraction, last_cement_fraction])

    fractions[0][0] = 0

    cement = 0  # cement layer fraction for the first casing (the deepest)
    pipe = 0
    fractions_per_section = [[pipe, cement]]
    for x in fractions:
        cement += x[1]
        pipe += x[0]
        fractions_per_section.append([pipe, cement])

    return fractions_per_section


def get_material(section, md, operation='drilling', first_casing_depth=None, water_depth=0):
    if first_casing_depth is None:
        first_casing_depth = -1

    if md <= first_casing_depth:
        if md >= water_depth:
            section_3 = 'casing'
        else:
            section_3 = 'riser'
    else:
        section_3 = 'formation'

    if md >= water_depth:
        section_4 = 'mixture'
        if first_casing_depth != -1:
            section_4 = 'mixture'
    else:
        section_4 = 'seawater'

    if operation == 'drilling':
        component = {'section_0': 'fluid',
                     'section_1': 'pipe',
                     'section_2': 'fluid',
                     'section_3': section_3,
                     'section_4': section_4}
    else:
        component = None
    return component[section]


def extra_calcs(well):
    h1, h2, h3, h3r = [], [], [], []
    sections = well.sections.copy()
    for x, y in zip(sections[0], sections[2]):
        # Reynolds number
        x['re'] = calc_re(well, x['rho'], x['visc'], section=0)  # Section 0
        y['re'] = calc_re(well, y['rho'], y['visc'], section=2)  # Section 2 - Annulus
        # Prandtl number
        x['pr'] = calc_pr(x['shc'], x['tc'], x['visc'])  # Section 0
        y['pr'] = calc_pr(y['shc'], y['tc'], y['visc'])  # Section 2 - Annulus
        # Friction factor
        x['f'] = calc_friction_factor(x['re'])  # Section 0
        y['f'] = calc_friction_factor(y['re'])  # Section 2 - Annulus
        # Nusselt number
        nu_int = calc_nu(well, x['f'], x['re'], x['pr'], section=1)  # Section 1
        nu_ext = calc_nu(well, y['f'], y['re'], y['pr'], section=1)  # Section 1
        nu_ann = calc_nu(well, y['f'], y['re'], y['pr'], section=2)  # Section 2 - Annulus
        # Heat transfer coefficients
        h1.append(calc_heat_transfer_coef(x['tc'], nu_int, well.pipe_id))
        h2.append(calc_heat_transfer_coef(x['tc'], nu_ext, well.pipe_od))
        h3.append(calc_heat_transfer_coef(x['tc'], nu_ann, well.annular_or * 2))
        h3r.append(calc_heat_transfer_coef(x['tc'], nu_ann, well.riser_ir * 2))

    return sections, h1, h2, h3, h3r


def calc_re(well, rho_fluid, viscosity, section=0):
    if section == 0:
        # Reynolds number inside pipe
        re = rho_fluid * well.vp * 2 * well.r1 / viscosity
    else:
        # Reynolds number - annular
        re = rho_fluid * well.va * 2 * (well.annular_or - well.r2) / viscosity

    return re


def calc_pr(shc_fluid, tc_fluid, viscosity):
    pr = viscosity * shc_fluid / tc_fluid

    return pr


def calc_friction_factor(re):
    f = 64 / re

    if 2300 < re < 10000:
        f = 1.63 / log(6.9 / re) ** 2

    if re >= 10000:
        f = 1.63 / log(6.9 / re) ** 2

    return f


def calc_nu(well, f, re, pr, section=1):
    nu = 4.36
    if section != 1 and re < 2300:
        nu = 1.86 * ((re * pr) ** (1 / 3)) * \
             ((2 * (well.annular_or - well.r2) / well.md[-1]) ** (1 / 3)) \
             * (1 ** 0.14)

    if 2300 < re < 10000:
        nu = (f / 8) * (re - 1000) * pr / (1 + (12.7 * (f / 8) ** 0.5) * (pr ** (2 / 3) - 1))

    if re >= 10000:
        nu = 0.027 * (re ** (4 / 5)) * (pr ** (1 / 3)) * (1 ** 0.14)

    return nu


def calc_heat_transfer_coef(tc, nu, diameter):
    h = tc * nu / diameter

    return h


def calc_formation_temp(well):
    temp_fm = [well.temp_surface]
    for j in range(1, well.cells_no):

        if j <= well.riser_cells:
            th_grad = well.th_grad_seawater  # Water Thermal Gradient for the Riser section
        else:
            th_grad = well.th_grad_fm  # Geothermal Gradient below the Riser section
        temp_delta = temp_fm[j - 1] + th_grad * (well.tvd[j] - well.tvd[j - 1])

        # Generating the Temperature Profile at t=0
        temp_fm.append(temp_delta)

    for x in well.sections:
        for y in range(well.cells_no):
            x[y]['temp'] = temp_fm[y]
            x[y]['temp_fm'] = temp_fm[y]

    return well.sections, temp_fm



def heat_coef(well, deltat, tt, t3):
    """
    Calculate heat transfer coefficients for each cell.
    :param t3: current temperature profile at section 3 (first casing)
    :param tt: current temperature profile at tubing wall
    :param well: a well object created from the function set_well()
    :param deltat: duration of each time step (seconds)
    :return: list with distribution of heat transfer coefficients
    """

    from math import pi, log
    from numpy import interp

    sections = [well.wd]
    if len(well.casings) > 0 and well.casings[0, 2] > 0:
        for i in range(len(well.casings))[::-1]:
            sections.append(well.casings[i, 2])

    vb = well.q / (pi * well.r3 ** 2)
    cbz = ((well.rhof[-1] * well.cf * vb) / well.deltaz) / 2  # Vertical component (North-South)
    cbe = (2 * well.h1[-1] / well.r3) / 2  # East component
    cbt = well.rhof[-1] * well.cf / deltat  # Time component

    # Creating empty lists

    # Section 1: Fluid in Tubing
    c1z = []
    c1e = []
    c1 = []
    c1t = []

    # Section 2: Tubing Wall
    c2z = []
    c2e = []
    c2w = []
    c2t = []

    # Section 3: Fluid in Annulus
    c3z = []
    c3e = []
    c3w = []
    c3t = []

    # Section 4: First casing
    c4z = []
    c4e = []
    c4w = []
    c4t = []

    # Section 5: Surrounding Space
    c5z = []
    c5e = []
    c5w = []
    c5t = []

    in_section = 1
    section_checkpoint = sections[0]

    for x in range(well.zstep):
        if x*well.deltaz >= section_checkpoint and in_section < len(sections)+1:
            in_section += 1
            if section_checkpoint != sections[-1]:
                section_checkpoint = sections[in_section-1]

        gr_t = 9.81 * well.alpha * abs((tt[x] - t3[x])) * (well.rhof[x] ** 2) * (well.dti ** 3) / (well.visc_a[x] ** 2)
        gr_c = 9.81 * well.alpha * abs((tt[x] - t3[x])) * (well.rhof[x] ** 2) * (((well.r3 - well.r2) * 2) ** 3) / (
                    well.visc_a[x] ** 2)
        ra_t = gr_t * well.pr_a[x]
        ra_c = gr_c * well.pr_a[x]
        inc = [0, 30, 45, 60, 90]
        c_base = [0.069, 0.065, 0.059, 0.057, 0.049]
        c = interp(well.inclination[x], inc, c_base, right=0.049)
        nu_a_t = c * (ra_t ** (1/3)) * (well.pr_a[x] ** 0.074)
        nu_a_c = c * (ra_c ** (1/3)) * (well.pr_a[x] ** 0.074)
        h2 = well.lambdaf_a * nu_a_t / (well.r2 * log(well.r3/well.r2))
        h3 = well.lambdaf_a * nu_a_c / (well.r2 * log(well.r3/well.r2))
        h3r = h3
        lambdal_eq = well.lambdaf_a * nu_a_t

        # fluid inside tubing
        qp = 0.2 * well.q * 2 * (well.f_p[x] * well.rhof[x] * (well.vp ** 2) *
                                 (well.md[-1] / (well.dti * 127.094 * 10 ** 6)))

        c1z.append(((well.rhof[x] * well.cf * well.vp) / well.deltaz) / 2)  # Vertical component (North-South)
        c1e.append((2 * well.h1[x] / well.r1) / 2)  # East component
        c1.append(qp / (pi * (well.r1 ** 2)))  # Heat source term
        c1t.append(well.rhof[x] * well.cf / deltat)  # Time component

        # tubing wall
        c2z.append((well.lambdat / (well.deltaz ** 2)) / 2)  # Vertical component (North-South)
        c2e.append((2 * well.r2 * h2 / ((well.r2 ** 2) - (well.r1 ** 2))) / 2)  # East component
        c2w.append((2 * well.r1 * well.h1[x] / ((well.r2 ** 2) - (well.r1 ** 2))) / 2)  # West component
        c2t.append(well.rhot * well.ct / deltat)  # Time component

        if in_section == 1:
            lambda4 = well.lambdar  # Thermal conductivity of the casing (riser in this section)
            lambda5 = well.lambdaw  # Thermal conductivity of the surrounding space (seawater)
            lambda45 = (lambda4 * (well.r4r - well.r3r) + lambda5 * (well.r5 - well.r4r)) / (
                    well.r5 - well.r3r)  # Comprehensive Thermal conductivity of the casing (riser) and
                                         # surrounding space (seawater)
            lambda56 = well.lambdaw  # Comprehensive Thermal conductivity of the surrounding space (seawater) and
                                     # formation (seawater)
            c4 = well.cr  # Specific Heat Capacity of the casing (riser)
            c5 = well.cw  # Specific Heat Capacity of the surrounding space (seawater)
            rho4 = well.rhor  # Density of the casing (riser)
            rho5 = well.rhow  # Density of the surrounding space (seawater)

            # fluid inside annular
            c3z.append((lambdal_eq / (well.deltaz ** 2)) / 2)  # Vertical component (North-South)
            c3e.append((2 * well.r3 * h3 / ((well.r3 ** 2) - (well.r2 ** 2))) / 2)  # East component
            c3w.append((2 * well.r2 * h2 / ((well.r3 ** 2) - (well.r2 ** 2))) / 2)  # West component
            c3t.append(well.rhof_a[x] * well.cf_a / deltat)  # Time component

        else:
            # fluid inside annular
            c3z.append((lambdal_eq / (well.deltaz ** 2)) / 2)  # Vertical component (North-South)
            c3e.append((2 * well.r3 * h3r / ((well.r3r ** 2) - (well.r2 ** 2))) / 2)  # East component
            c3w.append((2 * well.r2 * h2 / ((well.r3r ** 2) - (well.r2 ** 2))) / 2)  # West component
            c3t.append(well.rhof_a[x] * well.cf_a / deltat)  # Time component

        if 1 < in_section < len(sections):

            # calculation for surrounding space
            # thickness
            tcsr = 0
            tcem = 0
            for i in range(len(well.casings) - in_section):
                tcsr += (well.casings[i + 1, 0] - well.casings[i + 1, 1]) / 2
                tcem += (well.casings[i + 1, 1] - well.casings[i, 0]) / 2

                tcem += (well.casings[len(well.casings)-in_section+1, 1] -
                         well.casings[len(well.casings)-in_section, 0]) / 2
            if in_section == 2:
                tcem += (well.dsro - well.casings[-1, 0])
            xcsr = tcsr / (well.r5 - well.r4)  # fraction of surrounding space that is casing
            xcem = tcem / (well.r5 - well.r4)  # fraction of surrounding space that is cement
            xfm = 1 - xcsr - xcem  # fraction of surrounding space that is formation

            # thermal conductivity
            lambdasr = well.lambdac * xcsr + well.lambdacem * xcem + well.lambdafm * xfm
            lambdacsr = (well.lambdac * (well.r4 - well.r3) + lambdasr * (well.r5 - well.r4)) / (well.r5 - well.r3)
            lambdasrfm = (well.lambdac * (well.r5 - well.r4) + lambdasr * (well.rfm - well.r5)) / (well.rfm - well.r4)

            # Specific Heat Capacity
            csr = (well.cc * tcsr + well.ccem * tcem) / (well.r5 - well.r4)

            # Density
            rhosr = xcsr * well.rhoc + xcem * well.rhocem + xfm * well.rhofm

            lambda4 = well.lambdac
            lambda45 = lambdacsr
            lambda5 = lambdasr
            lambda56 = lambdasrfm
            c4 = well.cc  # Specific Heat Capacity of the casing
            c5 = csr  # Specific Heat Capacity of the surrounding space
            rho4 = well.rhoc  # Density of the casing
            rho5 = rhosr  # Density of the surrounding space

        if in_section == len(sections)+1:
            lambda4 = well.lambdafm
            lambda45 = well.lambdafm
            lambda5 = well.lambdafm
            lambda56 = well.lambdafm
            c4 = well.cfm  # Specific Heat Capacity of the casing (formation)
            c5 = well.cfm  # Specific Heat Capacity of the surrounding space (formation)
            rho4 = well.rhofm  # Density of the casing (formation)
            rho5 = well.rhofm  # Density of the surrounding space (formation)

        # first casing wall
        c4z.append((lambda4 / (well.deltaz ** 2)) / 2)
        c4e.append((2 * lambda45 / ((well.r4 ** 2) - (well.r3 ** 2))) / 2)
        c4w.append((2 * well.r3 * h3 / ((well.r4 ** 2) - (well.r3 ** 2))) / 2)
        c4t.append(rho4 * c4 / deltat)

        # surrounding space
        c5z.append((lambda5 / (well.deltaz ** 2)) / 2)
        c5w.append((lambda56 / (well.r5 * (well.r5 - well.r4) * log(well.r5 / well.r4))) / 2)
        c5e.append((lambda56 / (well.r5 * (well.r5 - well.r4) * log(well.rfm / well.r5))) / 2)
        c5t.append(rho5 * c5 / deltat)

    hc_1 = [c1z, c1e, c1, c1t]
    hc_2 = [c2z, c2e, c2w, c2t]
    hc_3 = [c3z, c3e, c3w, c3t]
    hc_4 = [c4z, c4e, c4w, c4t]
    hc_5 = [c5z, c5e, c5w, c5t]
    coefficients = [hc_1, hc_2, hc_3, hc_4, hc_5, cbe, cbt, cbz]

    return coefficients




import plotly.graph_objects as go
from numpy import polyfit, poly1d


def profile(temp_distribution, units='metric', operation='drilling'):

    pipe_name = {'drilling': 'Drill String',
                 'production': 'Production Tubing',
                 'injection': 'Injection Tubing'}

    # Plotting Temperature PROFILE

    fig = go.Figure()
    md = temp_distribution.md
    if units == 'english':
        md = [i * 3.28 for i in md]
    riser = temp_distribution.riser
    csg = temp_distribution.csgs_reach

    if operation == 'drilling':
        pipe_fluid = temp_distribution.tdsi

    else:
        pipe_fluid = temp_distribution.tft

    fig.add_trace(go.Scatter(x=pipe_fluid, y=md,
                             mode='lines',
                             name='Fluid in ' + pipe_name[operation]))
    fig.add_trace(go.Scatter(x=temp_distribution.ta, y=md,
                             mode='lines',
                             name='Fluid in Annulus'))

    if riser > 0:
        fig.add_trace(go.Scatter(x=temp_distribution.tr, y=md,
                                 mode='lines',
                                 name='Riser'))
    if csg > 0:
        fig.add_trace(go.Scatter(x=temp_distribution.tcsg, y=md,
                                 mode='lines',
                                 name='Casing'))
    fig.add_trace(go.Scatter(x=temp_distribution.tfm, y=md,
                             mode='lines',
                             name='Formation'))  # Temp. due to gradient vs Depth

    if units == 'metric':
        fig.update_layout(
            xaxis_title='Temperature, °C',
            yaxis_title='Depth, m')
    else:
        fig.update_layout(
            xaxis_title='Temperature, °F',
            yaxis_title='Depth, ft')

    title = 'Temperature Profile at %1.1f hours' % temp_distribution.time + ' of ' + operation
    fig.update_layout(title=title)

    fig.update_yaxes(autorange="reversed")

    return fig


def behavior(behavior_obj, operation='drilling'):
    """
    Plotting Tbottom and Tout through time
    """

    fig = go.Figure()

    time = behavior_obj.time

    if behavior_obj.finaltime <= 10:
        poly_order = 2
    else:
        poly_order = 10

    if operation == 'drilling' or operation == 'injection':
        tbot_smooth = polyfit(time, behavior_obj.tbot, poly_order)
        tbot = poly1d(tbot_smooth)(time)
        fig.add_trace(go.Scatter(x=time, y=tbot,
                                 mode='lines',
                                 name='Bottom'))    # Temp. Bottom vs Time

    if operation == 'drilling' or operation == 'production':
        tout_smooth = polyfit(time, behavior_obj.tout, poly_order)
        tout = poly1d(tout_smooth)(time)
        fig.add_trace(go.Scatter(x=time, y=tout,
                                 mode='lines',
                                 name='Outlet'))    # Temp. Oulet vs Time

    fig.add_trace(go.Scatter(x=time, y=[behavior_obj.tfm[-1]]*len(time),
                             mode='lines',
                             name='Formation @ TD'))  # Formation Temp. vs Time

    fig.update_layout(
        xaxis_title='Time, h',
        yaxis_title='Temperature, °C')

    title = 'Temperature behavior (%1.1f hours)' % behavior_obj.finaltime + ' Operation: ' + operation
    fig.update_layout(title=title)

    return fig


def plot_distribution(temp_distribution, time=None):
    """
    Plot the well temperature distribution.

    Arguments:
        temp_distribution (obj): well object with temperature data
        time (int or float): set respective time in hours

    Returns:
        plotly figure
    """

    if time is None:
        time = temp_distribution.time

    operation = temp_distribution.op
    pipe_name = {'drilling': 'Drill String',
                 'circulating': 'Pipe',
                 'production': 'Production Tubing',
                 'injection': 'Injection Tubing'}

    # Plotting Temperature PROFILE

    fig = go.Figure()
    md = temp_distribution.temperatures['md']
    riser = temp_distribution.riser_cells
    csg = temp_distribution.casings[0, 2]

    fig.add_trace(go.Scatter(x=temp_distribution.temperatures['in_pipe'], y=md,
                             mode='lines',
                             name='Fluid in ' + pipe_name[operation]))

    fig.add_trace(go.Scatter(x=temp_distribution.temperatures['annulus'], y=md,
                             mode='lines',
                             name='Fluid in Annulus'))

    if riser > 0:
        fig.add_trace(go.Scatter(x=temp_distribution.temperatures['riser'], y=md,
                                 mode='lines',
                                 name='Riser'))
    if csg > 0:
        fig.add_trace(go.Scatter(x=temp_distribution.temperatures['casing'], y=md,
                                 mode='lines',
                                 name='Casing'))
    fig.add_trace(go.Scatter(x=temp_distribution.temperatures['formation'], y=md,
                             mode='lines',
                             name='Formation'))  # Temp. due to gradient vs Depth

    fig.update_layout(
        xaxis_title='Temperature, °C',
        yaxis_title='Depth, m')

    title = 'Temperature Profile at %1.1f hours' % time + ' of ' + operation
    fig.update_layout(title=title)

    fig.update_yaxes(autorange="reversed")

    return fig


def plot_behavior(temp_behavior, title=True):
    fig = go.Figure()
    fig.add_trace(go.Scatter(x=temp_behavior.time, y=temp_behavior.bottom, mode='lines', name='Bottom'))
    fig.add_trace(go.Scatter(x=temp_behavior.time, y=temp_behavior.outlet, mode='lines', name='Outlet - annulus'))
    fig.add_trace(go.Scatter(x=temp_behavior.time, y=temp_behavior.max, mode='lines', name='Max. temp'))
    fig.add_trace(go.Scatter(x=temp_behavior.time, y=temp_behavior.formation_td, mode='lines', name='Formation at TD'))
    fig.update_layout(
        xaxis_title='time, h',
        yaxis_title='Temperature, °C')
    if title:
        fig.update_layout(title=str(temp_behavior.time[-1]) + ' hours of operation')

    return fig



def define_coef(coefficients, zstep):
    """
    Retrieves respective heat transfer coefficients for certain depth point.
    :param coefficients: list with distribution of heat transfer coefficients
    :param zstep: depth step
    :return: values of heat coefficients for each section at the same depth
    """

    hc_1 = coefficients[0]
    c1z = hc_1[0][zstep]
    c1e = hc_1[1][zstep]
    c1 = hc_1[2][zstep]
    c1t = hc_1[3][zstep]

    hc_2 = coefficients[1]
    c2z = hc_2[0][zstep]
    c2e = hc_2[1][zstep]
    c2w = hc_2[2][zstep]
    c2t = hc_2[3][zstep]

    hc_3 = coefficients[2]
    c3z = hc_3[0][zstep]
    c3e = hc_3[1][zstep]
    c3w = hc_3[2][zstep]
    c3t = hc_3[3][zstep]

    hc_4 = coefficients[3]
    c4z = hc_4[0][zstep]
    c4e = hc_4[1][zstep]
    c4w = hc_4[2][zstep]
    c4t = hc_4[3][zstep]

    hc_5 = coefficients[4]
    c5z = hc_5[0][zstep]
    c5e = hc_5[1][zstep]
    c5w = hc_5[2][zstep]
    c5t = hc_5[3][zstep]

    cbe = coefficients[5]
    cbt = coefficients[6]
    cbz = coefficients[7]

    return c1z, c1e, c1, c1t, c2z, c2e, c2w, c2t, c3z, c3e, c3w, c3t, c4z, c4e, c4w, c4t, c5z, c5e, c5w, c5t, cbe, \
           cbt, cbz


def temp_calc(well, initcond, heatcoeff):
    """
    Build the penta-diagonal matrix and solve it to get the well temperature distribution.
    :param well: a well object created from the function set_well()
    :param initcond: object with initial temperature profiles
    :param heatcoeff: list with distribution of heat transfer coefficients
    :return: object with final well temperature distribution
    """

    from numpy import zeros, linalg

    Tft = [well.tin]
    Tt = []
    Ta = []
    tc = []
    Tsr = []
    xi = 5

    # Creating vectors N,W,C,E,S,B
    N = []
    W = []
    C = []
    E = []
    S = []
    B = []

    for j in range(well.zstep):
        c1z, c1e, c1, c1t, c2z, c2e, c2w, c2t, c3z, c3e, c3w, c3t, c4z, c4e, c4w, c4t, c5z, c5e, c5w, c5t, cbe, \
        cbt, cbz = define_coef(heatcoeff, j)
        for i in range(xi):

            if i == 0:  # Inside Tubing

                if j == 1:
                    W.append(0)
                    C.append(c1t + c1e + c1z)
                    E.append(-c1e)
                    S.append(0)
                    B.append(c1t * initcond.tfto[j]    # Center(t=0)
                             + c1       # Heat Source
                             + c1e * (initcond.tto[j] - initcond.tfto[j])     # East(t=0)
                             + c1z * (initcond.tfto[j - 1] - initcond.tfto[j])
                             + c1z * well.tin)       # N/S(t=0)

                if 1 < j < well.zstep - 1:
                    N.append(-c1z)
                    W.append(0)
                    C.append(c1t + c1e + c1z)
                    E.append(-c1e)
                    S.append(0)
                    B.append(c1t * initcond.tfto[j]    # Center(t=0)
                             + c1       # Heat Source
                             + c1e * (initcond.tto[j] - initcond.tfto[j])     # East(t=0)
                             + c1z * (initcond.tfto[j - 1] - initcond.tfto[j]))       # N/S(t=0)

                if j == well.zstep - 1:
                    N.append(-c1z)
                    W.append(0)
                    C.append(cbt + c1z)
                    E.append(0)
                    B.append(cbt * initcond.tfto[j]    # Center(t=0)
                             + c1z * (initcond.tfto[j - 1] - initcond.tfto[j]))  # N/S(t=0)

            if i == 1:  # Tubing wall

                if j == 0:
                    C.append(c2t + c2e + c2w + c2z)
                    E.append(-c2e)
                    S.append(-c2z)
                    B.append(c2t * initcond.tto[j]
                             + c2e * (initcond.tao[j] - initcond.tto[j])
                             + c2w * (initcond.tfto[j] - initcond.tto[j])
                             + c2z * (initcond.tto[j + 1] - initcond.tto[j])
                             + c2w * well.tin)

                if 0 < j < well.zstep - 1:
                    N.append(-c2z)
                    W.append(-c2w)
                    C.append(c2t + c2e + c2w + 2 * c2z)
                    E.append(-c2e)
                    S.append(-c2z)
                    B.append(c2t * initcond.tto[j]
                             + c2e * (initcond.tao[j] - initcond.tto[j])
                             + c2w * (initcond.tfto[j] - initcond.tto[j])
                             + c2z * (initcond.tto[j + 1] - initcond.tto[j])
                             + c2z * (initcond.tto[j - 1] - initcond.tto[j]))

                if j == well.zstep - 1:
                    N.append(-c2z)
                    W.append(0)
                    C.append(c2t + c2z)
                    E.append(0)
                    B.append(c2t * initcond.tto[j]
                             + c2z * (initcond.tto[j - 1] - initcond.tto[j]))

            if i == 2:  # Annular

                if j == 0:
                    W.append(-c3w)
                    C.append(c3t + c3e + c3w + c3z)
                    E.append(-c3e)
                    S.append(-c3z)
                    B.append(c3t * initcond.tao[j]
                             + c3e * (initcond.tco[j] - initcond.tao[j])
                             + c3w * (initcond.tto[j] - initcond.tao[j])
                             + c3z * (initcond.tao[j + 1] - initcond.tao[j]))

                if 0 < j < well.zstep - 1:
                    N.append(-c3z)
                    W.append(-c3w)
                    C.append(c3t + c3e + c3w + 2 * c3z)
                    E.append(-c3e)
                    S.append(-c3z)
                    B.append(c3t * initcond.tao[j]
                             + c3e * (initcond.tco[j] - initcond.tao[j])
                             + c3w * (initcond.tto[j] - initcond.tao[j])
                             + c3z * (initcond.tao[j + 1] - initcond.tao[j])
                             + c3z * (initcond.tao[j - 1] - initcond.tao[j]))

                if j == well.zstep - 1:
                    N.append(-c3z)
                    W.append(0)
                    C.append(c3t + c3e + c3z)
                    E.append(-c3e)
                    B.append(c3t * initcond.tao[j]
                             + c3e * (initcond.tco[j] - initcond.tao[j])
                             + c3z * (initcond.tao[j - 1] - initcond.tao[j]))

            if i == 3:  # Casing

                if j == 0:
                    W.append(-c4w)
                    C.append(c4t + c4e + c4w + c4z)
                    E.append(-c4e)
                    S.append(-c4z)
                    B.append(c4t * initcond.tco[j]   # Center(t=0)
                             + c4e * (initcond.tsro[j] - initcond.tco[j])  # East(t=0)
                             + c4w * (initcond.tao[j] - initcond.tco[j])  # West(t=0)
                             + c4z * (initcond.tco[j + 1] - initcond.tco[j]))   # N/S(t=0)

                if 0 < j < well.zstep - 1:
                    N.append(-c4z)
                    W.append(-c4w)
                    C.append(c4t + c4e + c4w + 2 * c4z)
                    E.append(-c4e)
                    S.append(-c4z)
                    B.append(c4t * initcond.tco[j]    # Center(t=0)
                             + c4e * (initcond.tsro[j] - initcond.tco[j])     # East(t=0)
                             + c4w * (initcond.tao[j] - initcond.tco[j])      # West(t=0)
                             + c4z * (initcond.tco[j + 1] - 2 * initcond.tco[j] + initcond.tco[j - 1])) # N/S(t=0)

                if j == well.zstep - 1:
                    N.append(-c4z)
                    W.append(-c4w)
                    C.append(c4t + c4e + c4w + c4z)
                    E.append(-c4e)
                    B.append(c4t * initcond.tco[j]    # Center(t=0)
                             + c4e * (initcond.tsro[j] - initcond.tco[j])     # East(t=0)
                             + c4w * (initcond.tao[j] - initcond.tco[j])      # West(t=0)
                             + c4z * (initcond.tco[j - 1] - initcond.tco[j]))       # N/S(t=0))

            if i == 4:  # Surrounding Space

                if j == 0:
                    W.append(-c5w)
                    C.append(c5w + c5z + c5e + c5t)
                    E.append(0)
                    S.append(-c5z)
                    B.append(c5w * (initcond.tco[j] - initcond.tsro[j])
                             + c5z * (initcond.tsro[j + 1] - initcond.tsro[j])
                             + c5e * initcond.tsro[j])

                if 0 < j < well.zstep - 1:
                    N.append(-c5z)
                    W.append(-c5w)
                    C.append(c5w + c5e + 2 * c5z + c5t)
                    E.append(0)
                    S.append(-c5z)
                    B.append(c5w * (initcond.tco[j] - initcond.tsro[j])
                             + c5z * (initcond.tsro[j + 1] - initcond.tsro[j])
                             + c5z * (initcond.tsro[j - 1] - initcond.tsro[j])
                             + c5e * initcond.tsro[j])

                if j == well.zstep - 1:
                    N.append(-c5z)
                    W.append(-c5w)
                    C.append(c5w + c5e + c5z + c5t)
                    B.append(c5w * (initcond.tco[j] - initcond.tsro[j])
                             + c5z * (initcond.tsro[j - 1] - initcond.tsro[j])
                             + c5t * initcond.tsro[j]
                             + c5e * initcond.tsro[j])

    #LINEARSYSTEM
    # Creating pentadiagonal matrix
    A = zeros((xi * well.zstep - 1, xi * well.zstep - 1))

    # Filling up Pentadiagonal Matrix A
    lenC = xi * well.zstep - 1
    lenN = lenC - xi
    lenW = lenC - 1
    lenE = lenC - 1
    lenS = lenC - xi

    for it in range(lenC):  # Inserting list C
        A[it, it] = C[it]
    for it in range(lenE):  # Inserting list E
        A[it, it + 1] = E[it]
    for it in range(lenW):  # Inserting list W
        A[it + 1, it] = W[it]
    for it in range(lenN):  # Inserting list N
        A[it + xi, it] = N[it]
    for it in range(lenS):  # Inserting list S
        A[it, it + xi] = S[it]

    Temp = linalg.solve(A, B)

    for x in range(well.zstep):
        Tt.append(Temp[5 * x])
        if x < well.zstep-1:
            Tft.append(Temp[5 * x + 4])
            Ta.append(Temp[5 * x + 1])
        if x == well.zstep - 1:
            Ta.append(Tft[-1])
        tc.append(Temp[5 * x + 2])
        Tsr.append(Temp[5 * x + 3])


    t3 = tc.copy()

    tr = tc[:well.riser] + [None] * (well.zstep - well.riser)
    for x in range(well.riser):
        tc[x] = None

    csgs_reach = int(well.casings[0, 2] / well.deltaz)  # final depth still covered with casing(s)

    Toh = [None] * csgs_reach + tc[csgs_reach:]
    for x in range(csgs_reach, well.zstep):
        tc[x] = None

    class TempCalc(object):
        def __init__(self):
            self.tft = Tft
            self.tt = Tt
            self.ta = Ta
            self.t3 = t3
            self.tc = tc
            self.tr = tr
            self.tsr = Tsr
            self.toh = Toh
            self.csgs_reach = csgs_reach

    return TempCalc()



import numpy as np
from .heat_coefficients import add_heat_coefficients


def define_system_section0(well, sections, bit_position):
    for x, y in enumerate(sections[0][:bit_position+1]):
        if x == 0:
            y['N'] = 0
            y['W'] = 0
            y['C'] = 0
            y['E'] = 0
            y['S'] = 0
            y['B'] = 0
            y['temp'] = well.temp_inlet

        if x == 1:
            y['N'] = 0
            y['W'] = 0
            y['C'] = y['comp_time'] + y['comp_E'] + y['comp_N/S']
            y['E'] = - y['comp_E']
            y['S'] = 0
            y['B'] = y['comp_time'] * y['temp'] \
                + y['comp_HeatSource'] \
                + y['comp_E'] * (sections[1][x]['temp'] - y['temp']) \
                + y['comp_N/S'] * (sections[0][x - 1]['temp'] - y['temp']) \
                + y['comp_N/S'] * (sections[0][x - 1]['temp'])

        if 1 < x <= bit_position:
            y['N'] = - y['comp_N/S']
            y['W'] = 0
            y['C'] = y['comp_time'] + y['comp_E'] + y['comp_N/S']
            y['E'] = - y['comp_E']
            y['S'] = 0
            y['B'] = y['comp_time'] * y['temp'] \
                + y['comp_HeatSource'] \
                + y['comp_E'] * (sections[1][x]['temp'] - y['temp']) \
                + y['comp_N/S'] * (sections[0][x - 1]['temp'] - y['temp'])

    return well


def define_system_section1(well, sections, bit_position):
    for x, y in enumerate(sections[1][:bit_position+1]):
        if x == 0:
            y['N'] = 0
            y['W'] = 0
            y['C'] = y['comp_time'] + y['comp_E'] + y['comp_W'] + y['comp_N/S']
            y['E'] = - y['comp_E']
            y['S'] = - y['comp_N/S']
            y['B'] = y['comp_time'] * y['temp'] \
                + y['comp_E'] * (sections[2][x]['temp'] - y['temp']) \
                + y['comp_W'] * (sections[0][x]['temp'] - y['temp']) \
                + y['comp_W'] * sections[0][x]['temp'] \
                + y['comp_N/S'] * (sections[1][x + 1]['temp'] - y['temp'])

        if 0 < x < bit_position:
            y['N'] = - y['comp_N/S']
            y['W'] = - y['comp_W']
            y['C'] = y['comp_time'] + y['comp_E'] + y['comp_W'] + 2 * y['comp_N/S']
            y['E'] = - y['comp_E']
            y['S'] = - y['comp_N/S']
            y['B'] = y['comp_time'] * y['temp'] \
                + y['comp_E'] * (sections[2][x]['temp'] - y['temp']) \
                + y['comp_W'] * (sections[0][x]['temp'] - y['temp']) \
                + y['comp_N/S'] * (sections[1][x - 1]['temp'] - y['temp']) \
                + y['comp_N/S'] * (sections[1][x + 1]['temp'] - y['temp'])

        if x == bit_position:
            y['N'] = 0
            y['W'] = 0
            y['C'] = 0
            y['E'] = 0
            y['S'] = 0
            y['B'] = 0

    return well


def define_system_section2(well, sections, bit_position):
    for x, y in enumerate(sections[2][:bit_position+1]):
        if x < well.cells_no - 1:
            y['N'] = 0
            y['W'] = - y['comp_W']
            y['C'] = y['comp_time'] + y['comp_E'] + y['comp_W'] + y['comp_N/S']
            y['E'] = - y['comp_E']
            y['S'] = - y['comp_N/S']
            y['B'] = y['comp_time'] * y['temp'] \
                + y['comp_HeatSource'] \
                + y['comp_E'] * (sections[3][x]['temp'] - y['temp']) \
                + y['comp_W'] * (sections[1][x]['temp'] - y['temp']) \
                + y['comp_N/S'] * (sections[2][x + 1]['temp'] - y['temp'])

        if x == bit_position:
            y['N'] = 0
            y['W'] = 0
            y['C'] = 0
            y['E'] = 0
            y['S'] = 0
            y['B'] = 0

    return well


def define_system_section3(well, sections, bit_position):
    for x, y in enumerate(sections[3][:bit_position+1]):
        if x == 0:
            y['N'] = 0
            y['W'] = - y['comp_W']
            y['C'] = y['comp_time'] + y['comp_E'] + y['comp_W'] + y['comp_N/S']
            y['E'] = - y['comp_E']
            y['S'] = - y['comp_N/S']
            y['B'] = y['comp_time'] * y['temp'] \
                + y['comp_E'] * (sections[4][x]['temp'] - y['temp']) \
                + y['comp_W'] * (sections[2][x]['temp'] - y['temp']) \
                + y['comp_N/S'] * (sections[3][x + 1]['temp'] - y['temp'])

        if 0 < x < bit_position:
            y['N'] = - y['comp_N/S']
            y['W'] = - y['comp_W']
            y['C'] = y['comp_time'] + y['comp_E'] + y['comp_W'] + 2 * y['comp_N/S']
            y['E'] = - y['comp_E']
            y['S'] = - y['comp_N/S']
            y['B'] = y['comp_time'] * y['temp'] \
                + y['comp_E'] * (sections[4][x]['temp'] - y['temp']) \
                + y['comp_W'] * (sections[2][x]['temp'] - y['temp']) \
                + y['comp_N/S'] * (sections[3][x - 1]['temp'] - y['temp']) \
                + y['comp_N/S'] * (sections[3][x + 1]['temp'] - y['temp'])

        if x == bit_position:
            y['N'] = - y['comp_N/S']
            y['W'] = - y['comp_W']
            y['C'] = y['comp_time'] + y['comp_E'] + y['comp_W'] + y['comp_N/S']
            y['E'] = - y['comp_E']
            y['S'] = 0
            y['B'] = y['comp_time'] * y['temp'] \
                + y['comp_E'] * (sections[4][x]['temp'] - y['temp']) \
                + y['comp_W'] * (sections[2][x]['temp'] - y['temp']) \
                + y['comp_N/S'] * (sections[3][x - 1]['temp'] - y['temp'])

    return well


def define_system_section4(well, sections, bit_position):
    for x, y in enumerate(sections[4][:bit_position+1]):
        if x == 0:
            y['N'] = 0
            y['W'] = - y['comp_W']
            y['C'] = y['comp_time'] + y['comp_E'] + y['comp_W'] + y['comp_N/S']
            y['E'] = 0
            y['S'] = - y['comp_N/S']
            y['B'] = y['comp_time'] * y['temp'] \
                + y['comp_E'] * y['temp_fm'] \
                + y['comp_W'] * (sections[3][x]['temp'] - y['temp']) \
                + y['comp_N/S'] * (sections[4][x + 1]['temp'] - y['temp'])

        if 0 < x < bit_position:
            y['N'] = - y['comp_N/S']
            y['W'] = - y['comp_W']
            y['C'] = y['comp_time'] + y['comp_E'] + y['comp_W'] + 2 * y['comp_N/S']
            y['E'] = 0
            y['S'] = - y['comp_N/S']
            y['B'] = y['comp_time'] * y['temp'] \
                + y['comp_E'] * y['temp_fm'] \
                + y['comp_W'] * (sections[3][x]['temp'] - y['temp']) \
                + y['comp_N/S'] * (sections[4][x - 1]['temp'] - y['temp']) \
                + y['comp_N/S'] * (sections[4][x + 1]['temp'] - y['temp'])

        if x == bit_position:
            y['N'] = - y['comp_N/S']
            y['W'] = - y['comp_W']
            y['C'] = y['comp_time'] + y['comp_E'] + y['comp_W'] + y['comp_N/S']
            y['E'] = 0
            y['S'] = 0
            y['B'] = y['comp_time'] * y['temp'] \
                + y['comp_E'] * y['temp_fm'] \
                + y['comp_W'] * (sections[3][x]['temp'] - y['temp']) \
                + y['comp_N/S'] * (sections[4][x - 1]['temp'] - y['temp'])

    return well


def solve_pentadiagonal_system(well, bit_position):
    # Creating penta-diagonal matrix
    number_of_cells = bit_position + 1
    a = np.zeros((5 * number_of_cells, 5 * number_of_cells + 10))

    matrix = populate_matrix(a, well, bit_position)

    matrix = crop_matrix(matrix)

    constant_values = define_b_list(well, bit_position)

    temp_list = np.linalg.solve(matrix, constant_values)

    return temp_list


def populate_matrix(matrix, well, bit_position):
    row = 0
    column_base = 0

    for x in range(bit_position + 1):
        for y in well.sections:
            matrix[row, column_base] = y[x]['N']
            matrix[row, column_base + 4] = y[x]['W']
            matrix[row, column_base + 5] = y[x]['C']
            matrix[row, column_base + 6] = y[x]['E']
            matrix[row, column_base + 10] = y[x]['S']
            row += 1
            column_base += 1

    """# TRY THE CODE BELOW TO CHECK THAT THERE IS NOT ANY ROW WITH ONLY 0
    import pandas as pd
    df = pd.DataFrame(matrix)
    #df["sum"] = df.sum(axis=1)
    print(df)
    #print(df[df["sum"] == 0.0])"""

    return matrix


def crop_matrix(matrix):
    matrix = np.delete(matrix, 0, axis=0)
    matrix = np.delete(matrix, range(6), axis=1)
    matrix = np.delete(matrix, [-1, -2, -3, -4, -5], axis=1)
    matrix = np.delete(matrix, [-1, -2], axis=0)
    matrix = np.delete(matrix, [-1, -2], axis=1)

    matrix[-7, -3] = matrix[-7, -2]
    matrix[-7, -2] = 0
    matrix[-6, -3] = matrix[-6, -1]
    matrix[-6, -1] = 0

    """# Uncomment this to check the matrix cropped as a dataframe
    import pandas as pd
    df = pd.DataFrame(matrix)
    print(df.to_string())"""

    return matrix


def define_b_list(well, bit_position):
    b_list = []

    for x in range(bit_position + 1):
        for y in well.sections:
            b_list.append(y[x]['B'])

    b_list = b_list[1:-2]

    return b_list


def calc_temperature_distribution(well, time_step, bit_position):
    add_heat_coefficients(well, time_step, bit_position)
    well = add_values(well, bit_position)
    temp_list = solve_pentadiagonal_system(well, bit_position)
    well = update_temp(well, temp_list, bit_position)

    return well


def add_values(well, bit_position):
    well = define_system_section0(well, well.sections, bit_position)  # System section 0

    well = define_system_section1(well, well.sections, bit_position)  # System section 1

    well = define_system_section2(well, well.sections, bit_position)  # System section 2

    well = define_system_section3(well, well.sections, bit_position)  # System section 3

    well = define_system_section4(well, well.sections, bit_position)  # System section 4

    for x in ['N', 'W', 'C', 'E', 'S', 'B']:
        well.sections[1][bit_position][x] = well.sections[3][bit_position][x]
        well.sections[2][bit_position][x] = well.sections[4][bit_position][x]

    return well


def update_temp(well, temp_list, bit_position):
    rebuilt = [well.temp_inlet] + list(temp_list[:-3]) + [temp_list[-3]] * 3 + list(temp_list[-2:])

    list_index = 0
    for x in range(bit_position + 1):
        for y in well.sections:
            y[x]['temp'] = rebuilt[list_index]
            list_index += 1

    return well



from setuptools import setup
with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='pwptemp',
    packages=['pwptemp', 'pwptemp.production', 'pwptemp.injection'],
    version='0.3.5',
    license='LGPL v3',
    description='Well Temperature Distribution',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='Pro Well Plan AS',
    author_email='juan@prowellplan.com',
    url='https://github.com/pro-well-plan/pwptemp',
    keywords='well temperature distribution',
    classifiers=['Programming Language :: Python :: 3',
                 'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
                 'Natural Language :: English',
                 'Topic :: Scientific/Engineering',
                 'Topic :: Software Development',
                 'Topic :: Software Development :: Libraries',
                 'Topic :: Utilities'],
    install_requires=['numpy', 'plotly', 'torque_drag', 'scipy']
)



import numpy as np


def inputs_dict(casings=None):
    inputs = {'temp_inlet': None,           # fluid inlet temperature, °C
              'temp_surface': 15.0,         # surface temperature, °C
              'water_depth': 0.0,           # water depth, m
              'pipe_id': 4.0,               # pipe inner diameter, in
              'pipe_od': 4.5,               # pipe outer diameter, in
              'riser_id': 17.716,           # riser inner diameter, in
              'riser_od': 21.0,             # riser outer diameter, in
              'fm_diam': 80.0,              # undisturbed formation diameter, in
              'flowrate': 0.79,             # flow rate, m3/min
              'time': 10,                   # operation time if circulating, h

              # thermal conductivities, W / (m *°C)
              'tc_fluid': 0.635,            # fluid in pipe
              'tc_csg': 43.3,               # casing wall
              'tc_cem': 0.7,                # cement
              'tc_pipe': 40.0,              # pipe wall
              'tc_fm': 2.249,               # formation
              'tc_riser': 15.49,            # riser wall
              'tc_seawater': 0.6,           # seawater

              # specific heat capacities, J / (kg *°C)
              'shc_fluid': 3713.0,          # fluid in pipe
              'shc_csg': 469.0,             # casing wall
              'shc_cem': 2000.0,            # cement
              'shc_pipe': 400.0,            # pipe wall
              'shc_riser': 464.0,           # riser wall
              'shc_seawater': 4000.0,       # seawater
              'shc_fm': 800.0,              # formation

              # densities, sg
              'rho_fluid': 1.198,           # fluid in pipe
              'rho_pipe': 7.8,              # pipe wall
              'rho_csg': 7.8,               # casing wall
              'rho_riser': 7.8,             # riser wall
              'rho_fm': 2.245,              # formation
              'rho_seawater': 1.029,        # seawater
              'rho_cem': 2.7,               # cement

              'th_grad_fm': 0.0238,         # geothermal gradient, °C/m
              'th_grad_seawater': -0.005,   # seawater thermal gradient, °C/m
              'hole_diam': 0.216,           # diameter of open hole section, m
              'rpm': 100.0,                 # revolutions per minute
              'tbit': 5.0,                  # torque on the bit, kN*m
              'wob': 90.0,                  # weight on bit, kN
              'rop': [30.4],                # rate of penetration, m/h
              'an': 3100.0,                 # area of the nozzles, in^2
              'bit_n': 1.0,                 # drill bit efficiency, 0 to 1
              'dp_e': 0.0,                  # drill pipe eccentricity
              'thao_o': 1.82,               # yield stress, Pa
              'beta': 44983 * 10 ** 5,      # isothermal bulk modulus, Pa
              'alpha': 960 * 10 ** -6,      # expansion coefficient, 1/°C
              'k': 0.3832,                  # consistency index, Pa*s^n
              'n': 0.7,                     # flow behavior index, dimensionless
              'visc': 0.009                 # fluid viscosity
              }

    dict_with_casings = add_casings(casings, inputs)

    return dict_with_casings


def add_casings(casings, inputs):
    if casings is None:
        inputs['casings'] = [[(inputs['hole_diam'] + inputs['riser_od'] * 0.0254), inputs['hole_diam'], 0]]
        inputs['casings'] = np.asarray(inputs['casings'])
    else:
        csg_od = sorted([x['od'] * 0.0254 for x in casings])  # in to m
        csg_id = sorted([x['id'] * 0.0254 for x in casings])  # in to m
        depth = sorted([x['depth'] for x in casings], reverse=True)
        inputs['casings'] = [[csg_od[x], csg_id[x], depth[x]] for x in range(len(casings))]
        inputs['casings'] = np.asarray(inputs['casings'])

    return inputs



def data(casings=[], d_openhole=0.216, units='metric'):
    """
    Parameters involved within the operation calculations
    :param casings: list of dictionaries with casings characteristics (od, id and depth)
    :param d_openhole: diameter of open hole section, m
    :param units: system of units ('metric' or 'english')
    :return: a dictionary with default values for the required parameters
    """

    from numpy import asarray

    dict_met = {'ts': 15.0, 'wd': 100.0,  'dti': 4.0, 'dto': 4.5, 'dri': 17.716, 'dro': 21.0, 'dfm': 80.0,
                'q': 2000, 'lambdaf': 0.635, 'lambdaf_a': 0.635, 'lambdac': 43.3, 'lambdacem': 0.7, 'lambdat': 40.0,
                'lambdafm': 2.249, 'lambdar': 15.49, 'lambdaw': 0.6, 'cf': 3713.0, 'cf_a': 3713.0, 'cc': 469.0,
                'ccem': 2000.0, 'ct': 400.0, 'cr': 464.0, 'cw': 4000.0, 'cfm': 800.0, 'rhof': 0.85, 'rhof_a': 1.2,
                'rhot': 7.8, 'rhoc': 7.8, 'rhor': 7.8, 'rhofm': 2.245, 'rhow': 1.029, 'rhocem': 2.7, 'gt': 0.0238,
                'wtg': -0.005, 'api': 30, 'beta': 44983 * 10 ** 5, 'alpha': 960 * 10 ** -6, 'beta_a': 44983 * 10 ** 5,
                'alpha_a': 960 * 10 ** -6}

    dict_eng = {'ts': 59.0, 'wd': 328.0, 'dti': 4.0, 'dto': 4.5, 'dri': 17.716, 'dro': 21.0, 'dfm': 80.0,
                'q': 366.91, 'lambdaf': 1.098, 'lambdaf_a': 1.098, 'lambdac': 74.909, 'lambdacem': 1.21,
                'lambdat': 69.2, 'lambdafm': 3.89, 'lambdar': 26.8, 'lambdaw': 1.038, 'cf': 0.887, 'cf_a': 0.887,
                'cc': 0.112, 'ccem': 0.478, 'ct': 0.096, 'cr': 0.1108, 'cw': 0.955, 'cfm': 0.19, 'rhof': 7.09,
                'rhof_a': 10, 'rhot': 65.09, 'rhoc': 65.09, 'rhor': 65.09, 'rhofm': 18.73, 'rhow': 8.587,
                'rhocem': 22.5, 'gt': 0.00403, 'wtg': -8.47*10**-4, 'api': 30, 'beta': 652423,
                'alpha': 5.33 * 10 ** -4, 'beta_a': 652423, 'alpha_a': 5.33 * 10 ** -4}

    if units == 'metric':
        dict = dict_met
    else:
        dict = dict_eng

    if len(casings) > 0:
        od = sorted([x['od'] * 0.0254 for x in casings])
        id = sorted([x['id'] * 0.0254 for x in casings])
        depth = sorted([x['depth'] for x in casings], reverse=True)
        dict['casings'] = [[od[x], id[x], depth[x]] for x in range(len(casings))]
        dict['casings'] = asarray(dict['casings'])
    else:
        dict['casings'] = [[(d_openhole + dict['dro'] * 0.0254), d_openhole, 0]]
        dict['casings'] = asarray(dict['casings'])

    return dict


def info(about='all'):
    """
    Retrieves information about the parameters (description and units)
    :param about: type of parameters
    :return: description and units of parameters
    """

    print("Use the ID of a parameter to change the default value (e.g. tdict['tin']=30 to change the fluid inlet "
          "temperature from the default value to 30° Celsius)")
    print('Notice that the information is provided as follows:' + '\n' +
          'parameter ID: general description, units' + '\n')

    tubular_parameters = 'VALUES RELATED TO TUBULAR SIZES' + '\n' + \
                         'dti: tubing inner diameter, in' + '\n' + \
                         'dto: tubing outer diameter, in' + '\n' + \
                         'dri: riser inner diameter, in' + '\n' + \
                         'dro: riser outer diameter, in' + '\n'

    conditions_parameters = 'PARAMETERS RELATED TO SIMULATION CONDITIONS' + '\n' + \
                            'ts: surface temperature, °C or °F' + '\n' + \
                            'wd: water depth, m or ft' + '\n' + \
                            'dfm: undisturbed formation diameter, m or ft' + '\n'

    heatcoeff_parameters = 'PARAMETERS RELATED TO HEAT COEFFICIENTS' + '\n' + \
                           'lambdaf: production fluid - thermal conductivity, W/(m*°C) or BTU/(h*ft*°F)' + '\n' + \
                           'lambdaf_a: annular fluid - thermal conductivity, W/(m*°C) or BTU/(h*ft*°F)' + '\n' + \
                           'lambdac: casing - thermal conductivity, W/(m*°C) or BTU/(h*ft*°F)' + '\n' + \
                           'lambdacem: cement - thermal conductivity, W/(m*°C) or BTU/(h*ft*°F)' + '\n' + \
                           'lambdat: tubing - thermal conductivity, W/(m*°C) or BTU/(h*ft*°F)' + '\n' + \
                           'lambdafm: formation - thermal conductivity, W/(m*°C) or BTU/(h*ft*°F)' + '\n' + \
                           'lambdar: riser - thermal conductivity, W/(m*°C) or BTU/(h*ft*°F)' + '\n' + \
                           'lambdaw: water - thermal conductivity, W/(m*°C) or BTU/(h*ft*°F)' + '\n' + \
                           'cf: production fluid - specific heat capacity, J/(kg*°C) or BTU/(lb*°F)' + '\n' + \
                           'cf_a: annular fluid - specific heat capacity, J/(kg*°C) or BTU/(lb*°F)' + '\n' + \
                           'cc: casing - specific heat capacity, J/(kg*°C) or BTU/(lb*°F)' + '\n' + \
                           'ccem: cement - specific heat capacity, J/(kg*°C) or BTU/(lb*°F)' + '\n' + \
                           'ct: tubing - specific heat capacity, J/(kg*°C) or BTU/(lb*°F)' + '\n' + \
                           'cr: riser - specific heat capacity, J/(kg*°C) or BTU/(lb*°F)' + '\n' + \
                           'cw: water - specific heat capacity, J/(kg*°C) or BTU/(lb*°F)' + '\n' + \
                           'cfm: formation - specific heat capacity, J/(kg*°C) or BTU/(lb*°F)' + '\n' + \
                           'gt: geothermal gradient, °C/m or °F/ft' + '\n' + \
                           'wtg: seawater thermal gradient, °C/m or °F/ft' + '\n'

    densities_parameters = 'PARAMETERS RELATED TO DENSITIES' + '\n' + \
                           'rhof: production fluid density, sg or ppg' + '\n' + \
                           'rhof_a: annular fluid density, sg or ppg' + '\n' + \
                           'rhot: tubing density, sg or ppg' + '\n' + \
                           'rhoc: casing density, sg or ppg' + '\n' + \
                           'rhor: riser density, sg or ppg' + '\n' + \
                           'rhofm: formation density, sg or ppg' + '\n' + \
                           'rhow: seawater density, sg or ppg' + '\n' + \
                           'rhocem: cement density, sg or ppg' + '\n' + \
                           'api: api gravity, °API' + '\n' + \
                           'beta: isothermal bulk modulus of production fluid, Pa' + '\n' + \
                           'alpha: expansion coefficient of production fluid, 1/°C' + '\n' + \
                           'beta_a: isothermal bulk modulus of fluid in annular, Pa' + '\n' + \
                           'alpha_a: expansion coefficient of fluid in annular, 1/°C or 1/°F' + '\n'


    operational_parameters = 'PARAMETERS RELATED TO THE OPERATION' + '\n' + \
                             'q: flow rate, m^3/day or bbl/day' + '\n'

    if about == 'casings':
        print(tubular_parameters)

    if about == 'conditions':
        print(conditions_parameters)

    if about == 'heatcoeff':
        print(heatcoeff_parameters)

    if about == 'densities':
        print(densities_parameters)

    if about == 'operational':
        print(operational_parameters)

    if about == 'all':
        print(tubular_parameters + '\n' + conditions_parameters + '\n' + heatcoeff_parameters + '\n' +
              densities_parameters + '\n' + operational_parameters)


def set_well(temp_dict, depths, units='metric'):
    """
    Define properly the parameters and respective values within an object well.
    :param temp_dict: dictionary with inputs and default values.
    :param depths: wellpath object
    :param units: system of units ('metric' or 'english')
    :return: a well object with conditions and parameters defined
    """

    from math import pi, log

    class NewWell(object):
        def __init__(self):
            # DEPTH
            self.md = depths.md
            self.tvd = depths.tvd
            self.deltaz = self.md[-1] - self.md[-2]
            self.zstep = depths.cells_no
            self.sections = depths.sections
            self.north = depths.north
            self.east = depths.east
            self.inclination = depths.inclination
            self.dogleg = depths.dogleg
            self.azimuth = depths.azimuth
            if units != 'metric':
                self.md = [i / 3.28 for i in self.md]
                self.tvd = [i / 3.28 for i in self.tvd]
                self.deltaz = self.deltaz / 3.28
                self.north = [i / 3.28 for i in self.north]
                self.east = [i / 3.28 for i in self.east]

            # TUBULAR
            if units == 'metric':
                d_conv = 0.0254   # from in to m
            else:
                d_conv = 0.0254   # from in to m
            self.casings = temp_dict["casings"]  # casings array
            self.dti = temp_dict["dti"] * d_conv  # Tubing Inner  Diameter, m
            self.dto = temp_dict["dto"] * d_conv   # Tubing Outer Diameter, m
            self.dri = temp_dict["dri"] * d_conv  # Riser diameter Inner Diameter, m
            self.dro = temp_dict["dro"] * d_conv   # Riser diameter Outer Diameter, m

            # CONDITIONS
            if units == 'metric':
                depth_conv = 1  # from m to m
                self.ts = temp_dict["ts"]  # Surface Temperature (RKB), °C
            else:
                depth_conv = 1 / 3.28  # from ft to m
                self.ts = (temp_dict["ts"] - 32) * (5 / 9)  # Surface Temperature (RKB), from °F to °C
            self.wd = temp_dict["wd"] * depth_conv  # Water Depth, m
            self.riser = round(self.wd / self.deltaz)  # number of grid cells for the riser
            self.dsr = self.casings[0, 0]  # Surrounding Space Inner Diameter, m
            self.dsro = sorted([self.dro + 0.03, self.casings[-1, 0] + 0.03])[-1]  # Surrounding Space Outer Diameter, m
            self.dfm = temp_dict["dfm"] * d_conv  # Undisturbed Formation Diameter, m

            # RADIUS (CALCULATED)
            self.r1 = self.dti / 2  # Tubing Inner  Radius, m
            self.r2 = self.dto / 2  # Tubing Outer Radius, m
            self.r3 = self.casings[0, 1] / 2  # Casing Inner Radius, m
            self.r3r = self.dri / 2  # Riser Inner Radius, m
            self.r4r = self.dro / 2  # Riser Outer Radius, m
            self.r4 = self.casings[0, 0] / 2  # Surrounding Space Inner Radius m
            self.r5 = self.dsro / 2  # Surrounding Space Outer Radius, m
            self.rfm = self.dfm / 2  # Undisturbed Formation Radius, m

            # DENSITIES kg/m3
            if units == 'metric':
                dens_conv = 1000     # from sg to kg/m3
            else:
                dens_conv = 119.83   # from ppg to kg/m3
            self.rhof = temp_dict["rhof"] * dens_conv  # Fluid
            self.rhof_a = temp_dict["rhof_a"] * dens_conv  # Fluid
            self.rhot = temp_dict["rhot"] * dens_conv  # Tubing
            self.rhoc = temp_dict["rhoc"] * dens_conv  # Casing
            self.rhor = temp_dict["rhor"] * dens_conv  # Riser
            self.rhocem = temp_dict["rhocem"] * dens_conv  # Cement Sheath
            self.rhofm = temp_dict["rhofm"] * dens_conv  # Formation
            self.rhow = temp_dict["rhow"] * dens_conv  # Seawater

            # OPERATIONAL
            if units == 'metric':
                q_conv = 0.04167     # from m^3/day to m^3/h
            else:
                q_conv = 0.0066244706   # from bbl/day to m^3/h
            self.q = temp_dict["q"] * q_conv  # Flow rate, m^3/h
            self.vp = (self.q / (pi * (self.r1 ** 2))) / 3600  # Fluid velocity through the tubing

            # HEAT COEFFICIENTS
            if units == 'metric':
                lambda_conv = 1     # from W/(m*°C) to W/(m*°C)
                c_conv = 1  # from J/(kg*°C) to J/(kg*°C)
                gt_conv = 1     # from °C/m to °C/m
                beta_conv = 1   # from Pa to Pa
                alpha_conv = 1  # from 1/°F to 1/°C
            else:
                lambda_conv = 1/1.73     # from BTU/(h*ft*°F) to W/(m*°C)
                c_conv = 4187.53  # from BTU/(lb*°F) to J/(kg*°C)
                gt_conv = 3.28*1.8     # from °F/ft to °C/m
                beta_conv = 6894.76  # from psi to Pa
                alpha_conv = 1.8  # from 1/°F to 1/°C

            # Thermal conductivity  W/(m*°C)
            self.lambdaf = temp_dict["lambdaf"] * lambda_conv  # Production Fluid
            self.lambdaf_a = temp_dict["lambdaf_a"] * lambda_conv  # Annular Fluid
            self.lambdac = temp_dict["lambdac"] * lambda_conv   # Casing
            self.lambdacem = temp_dict["lambdacem"] * lambda_conv   # Cement
            self.lambdat = temp_dict["lambdat"] * lambda_conv   # Tubing wall
            self.lambdafm = temp_dict["lambdafm"] * lambda_conv        # Formation
            self.lambdar = temp_dict["lambdar"] * lambda_conv      # Riser
            self.lambdaw = temp_dict["lambdaw"] * lambda_conv      # Seawater

            self.beta = temp_dict["beta"] * beta_conv       # isothermal bulk modulus in tubing, Pa
            self.alpha = temp_dict['alpha'] * alpha_conv    # Fluid Thermal Expansion Coefficient in tubing, 1/°C
            self.beta_a = temp_dict["beta_a"] * beta_conv        # isothermal bulk modulus in annular, Pa
            self.alpha_a = temp_dict['alpha_a'] * alpha_conv    # Fluid Thermal Expansion Coefficient in annular, 1/°C

            # Specific heat capacity, J/(kg*°C)
            self.cf = temp_dict["cf"] * c_conv       # Production Fluid
            self.cf_a = temp_dict["cf"] * c_conv  # Annular Fluid
            self.cc = temp_dict["cc"] * c_conv    # Casing
            self.ccem = temp_dict["ccem"] * c_conv     # Cement
            self.ct = temp_dict["ct"] * c_conv     # Tubing
            self.cr = temp_dict["cr"] * c_conv     # Riser
            self.cw = temp_dict["cw"] * c_conv      # Seawater
            self.cfm = temp_dict["cfm"] * c_conv       # Formation

            self.gt = temp_dict["gt"] * gt_conv * self.deltaz  # Geothermal gradient, °C/m
            self.wtg = temp_dict["wtg"] * gt_conv * self.deltaz  # Seawater thermal gradient, °C/m

            # Raise Errors:

            if self.casings[-1, 0] > self.dsro:
                raise ValueError('Last casing outer diameter must be smaller than the surrounding space diameter.')

            if self.casings[0, 2] > self.md[-1]:
                raise ValueError('MD must be higher than the first casing depth.')

            if self.casings[0, 1] < self.dto:
                raise ValueError('Tubing outer diameter must be smaller than the first casing inner diameter.')

            if self.wd > 0 and self.dro > self.dsro:
                raise ValueError('Riser diameter must be smaller than the surrounding space diameter.')

            if self.dsro > self.dfm:
                raise ValueError('Surrounding space diameter must be smaller than the undisturbed formation diameter.')

        def wellpath(self):
            """
            :return: wellpath object
            """
            return depths

        def define_density(self, ic, cond=0):
            """
            Calculate the density profile
            :param ic: current temperature distribution
            :param cond: '0' to calculate the initial profile
            :return: density profile and derived calculations
            """

            from .fluid import initial_density, calc_density

            if cond == 0:
                self.rhof, self.rhof_initial = initial_density(self, ic)
                self.rhof_a, self.rhof_a_initial = initial_density(self, ic, section='annular')
            else:
                self.rhof = calc_density(self, ic, self.rhof_initial)
                self.rhof_a = calc_density(self, ic, self.rhof_initial, section='annular')
            # Reynolds number inside tubing
            self.re_p = [x * self.vp * 2 * self.r1 / y for x, y in zip(self.rhof, self.visc_t)]
            self.f_p = []  # Friction factor inside tubing
            self.nu_dpi = []
            for x in range(len(self.md)):
                if self.re_p[x] < 2300:
                    self.f_p.append(64 / self.re_p[x])
                    self.nu_dpi.append(4.36)
                else:
                    self.f_p.append(1.63 / log(6.9 / self.re_p[x]) ** 2)
                    self.nu_dpi.append(
                        (self.f_p[x] / 8) * (self.re_p[x] - 1000) *
                        self.pr_t[x] / (1 + (12.7 * (self.f_p[x] / 8) ** 0.5) *
                        (self.pr_t[x] ** (2 / 3) - 1)))
            # convective heat transfer coefficients, W/(m^2*°C)
            self.h1 = [self.lambdaf * x / self.dti for x in self.nu_dpi]  # Tubing inner wall

            return self

        def define_viscosity(self, ic):
            """
            Calculate the viscosity profile
            :param ic: current temperature distribution
            :return: viscosity profile and derived calculations
            """

            from .fluid import calc_vicosity
            self.api = temp_dict["api"]
            self.visc_t, self.visc_a = calc_vicosity(self.api, ic)

            self.pr_t = [x * self.cf / self.lambdaf for x in self.visc_t]  # Prandtl number - inside tubing
            self.pr_a = [x * self.cf / self.lambdaf for x in self.visc_a]  # Prandtl number - annulus

            return self

    return NewWell()
